import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;

class GenCode implements Visitor<Void> {
	String dateiname;
	
	
	public GenCode(String dateiname) {
		this.dateiname = dateiname;
	}

	Writer out;

	Map<String, Integer> env = Map.of();
	String[] rs = { "%rdi", "%rsi", "%rdx", "%rcx", "%r8", "%r9" };

	@Override
	public Void visit(Tree.Prog prog) {
		try {
			out = new FileWriter(dateiname+ ".s");
			prog.funs.stream().forEach(fun -> fun.welcome(this));

			out.write(".globl  start\n");
			out.write(".type   start, @function\n");
			out.write("start:\n");
			out.write("        pushq   %rbp\n");
			out.write("        movq    %rsp, %rbp\n");
			prog.expr.welcome(this);
			out.write("        popq    %rax\n");
			out.write("        popq    %rbp\n");
			out.write("        ret\n");
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Void visit(Tree.MultExpr e) {
		try {
			e.op1.welcome(this);
			e.op2.welcome(this);
			out.write("        popq    %rax\n");
			out.write("        popq    %rdx\n");
			out.write("        imulq   %rdx, %rax\n");
			out.write("        pushq   %rax\n");
		} catch (Exception ex) {
		}

		return null;
	}

	@Override
	public Void visit(Tree.AddExpr e) {
		try {

			e.op1.welcome(this);
			e.op2.welcome(this);
			out.write("        popq    %rax\n");
			out.write("        popq    %rdx\n");
			out.write("        addq   %rdx, %rax\n");
			out.write("        pushq   %rax\n");
		} catch (Exception ex) {
		}

		return null;
	}

	@Override
	public Void visit(Tree.IntLit e) {
		try {
			out.write("	    pushq    $" + e.n + "\n");
		} catch (Exception ex) {
		}
		return null;
	}

	@Override
	public Void visit(Tree.Var v) {
		try {
			out.write("        movq "+env.get(v.name)+"(%rbp), %rax\n");
			out.write("        pushq %rax\n");
		} catch (IOException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Void visit(Tree.FunDef e) {
		try {
			out.write(".globl  " + e.name + "\n");
			out.write(".type   " + e.name + ", @function\n");
			out.write(e.name + ":\n");
			out.write("        pushq   %rbp\n");
			out.write("        movq    %rsp, %rbp\n");

			var registerArgs = Math.min(rs.length, e.args.size());
			env = new HashMap<>();

			var sp = -8;
			for (var i = 0; i < registerArgs; i++) {
				out.write("        movq " + rs[i] + ", " + sp + "(%rbp)\n");
				env.put(e.args.get(i), sp);
				sp = sp - 8;
			}
			var stackChange = -sp;
			out.write("        subq $" + stackChange + ", %rsp\n");

			sp = 16;
			for (var i = registerArgs; i < e.args.size(); i++) {
				env.put(e.args.get(i), sp);
				sp += 8;
			}

			e.body.welcome(this);
			out.write("        popq    %rax\n");
			out.write("        addq $" + stackChange + ", %rsp\n");

			out.write("        popq    %rbp\n");
			out.write("        ret\n");
		} catch (java.io.IOException e3) {

		}
		return null;
	}

	@Override
	public Void visit(Tree.FunCall e) {
		try {
			for (int i = 0; i < Math.min(rs.length, e.args.size()); i++) {
				e.args.get(i).welcome(this);
				out.write("        popq " + rs[i] + "\n");
			}
			e.args.stream().skip(rs.length).forEach(arg -> {
				arg.welcome(this);
			});
			out.write("        call " + e.name + "\n");
			out.write("        pushq %rax\n");
		} catch (IOException e1) {

		}

		return null;
	}

	public static void main(String[] args) throws Exception {
		LParser parser = new LParser(new FileReader(args[0]));
		var gj = new GenCode(args[0].substring(0,args[0].indexOf('.')));
		System.out.println(parser.program().welcome(gj));

	}
}
